# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

import attrs

from .base import ButtonStyle, Component, ComponentType

if typing.TYPE_CHECKING:
    from wyvern.models.emojis import CustomEmoji

__all__: tuple[str, ...] = ("Button",)


@attrs.define(kw_only=True, slots=True, repr=True)
class Button(Component):
    """Represents a discord button.
    The properties mentioned below can be used to create a button.
    """

    style: ButtonStyle = ButtonStyle.SECONDARY
    """Style of the button."""
    label: str | None = None
    """Button's label."""
    emoji: str | "CustomEmoji" | None = None
    """Emoji embedded in the button."""
    custom_id: str | None = None
    """Custom id for the component."""
    disabled: bool = False
    """True if the component is disabled."""
    url: str | None = None
    """The URL this button points to, if any."""

    type: ComponentType = ComponentType.BUTTON

    def to_payload(self) -> dict[str, typing.Any]:
        payload: dict[str, typing.Any] = {
            "type": int(self.type),
            "style": int(self.style),
            "label": self.label,
            "custom_id": self.custom_id or "wyvern.NO_CUSTOM_ID",
            "disabled": self.disabled,
            "url": self.url,
        }
        if self.emoji is not None:
            payload["emoji"] = (
                {"name": self.emoji}
                if (not isinstance(self.emoji, CustomEmoji))
                else {"name": self.emoji.name, "id": self.emoji.id, "animated": self.emoji.is_animated}
            )
        return payload



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import enum
import typing

import attrs

from wyvern.utils import get_arg_count

if typing.TYPE_CHECKING:
    from wyvern._types import AnyCallableT, CheckT
    from wyvern.clients import GatewayClient


__all__: tuple[str, ...] = ("Event", "EventListener", "EventHandler", "as_listener")


@typing.final
class Event(enum.Enum):
    """Event Enums."""

    READY = "READY"

    MESSAGE_CREATE = "MESSAGE_CREATE"
    """Triggered when a message is created.
    
    Arguments provided:

    * message ([wyvern.Message][])
    """
    INTERACTION_CREATE = "INTERACTION_CREATE"
    """Triggered when an interaction is created.
    
    Arguments provided:

    * interaction ([wyvern.Interaction])
    
    !!! note
        The interaction argument can be any derivative of interaction!.
    """
    GUILD_CREATE = "GUILD_CREATE"
    # Library Events

    STARTING = "STARTING"
    """Triggered when bot is loaded and is starting.

    Note: This event gets triggerd once in a runtime.
    
    Arguments provided:
    
    * client ([wyvern.GatewayClient][])"""
    STARTED = "STARTED"
    """Triggered when the bot has successfully verified its token and is running.
    
    Arguments provided:
    
    * client ([wyvern.GatewayClient][])"""
    GATEWAY_CONNECTED = "GATEWAY_CONNECTED"
    """Triggered when the gateway initalises a connection.
    
    Arguments provided:

    * client ([wyvern.GatewayClient][])"""


@typing.final
@attrs.define
class EventListener:
    """Class representating a callable to be called when a specific event type is
    dispatched.

    Parameters
    ----------
    event_type : str | wyvern.events.Event
        The type of event this listener belongs to.
    callback : typing.Callable[..., typing.Awaitable[typing.Any]]
        The coroutine to run when event is dispatched
    max_trigger : int | float
        Max amount of time this listener will be triggered.
    checks: list[wyvern._types.CheckT]
        Checks added to the event listener.

    Attributes
    ----------
    trigger_count: int
        Number of times this listener has been triggered
    """

    __parent_identity__: str = attrs.field(init=False, default="-")
    event_type: Event
    callback: AnyCallableT
    max_trigger: int | float
    trigger_count: int = 0
    event_handler: EventHandler | None = None
    checks: list[CheckT] = attrs.field(init=False, default=[])

    def check(self, predicate: CheckT) -> CheckT:
        """Adds an check to the listener.
        The check callback accepts the same arguments as the related event.
        The predicate should return a bool ( strict to True or False ).

        Example
            import wyvern

            client = wyvern.GatewayClient("TOKEN")

            @client.with_listener(wyvern.Event.MESSAGE_CREATE)
            async def msg_create(msg: wyvern.Message) -> None:
                await msg.respond("This message was created when the check passed.")

            @msg_create.check
            async def msg_create_check(msg: wyvern.Message) -> bool:
                return msg.author.id == 1234567890123456789

            client.run()


        """

        def inner() -> CheckT:
            nonlocal predicate
            if not get_arg_count(predicate) == get_arg_count(self.callback):
                raise TypeError("EventListener and check callbacks should accept same number of arguments.")
            self.checks.append(predicate)
            return predicate

        return inner()

    async def process_checks(self, *args: typing.Any) -> bool:
        assert self.event_handler
        results = [await check(*args) for check in self.checks]
        if any(faulty_checks := [results.index(result) for result in results if result not in (True, False)]):
            self.event_handler.client._logger.warning(
                "Got non-bool return values from checks in %s: %s",
                self.__call__.__name__,
                ", ".join(map(lambda c: self.checks[c].__name__, faulty_checks)),
            )
        return all(results)

    async def __call__(self, *args: typing.Any) -> None:
        if (await self.process_checks(*args)) is False:
            return
        assert (handler := self.event_handler) is not None
        finalargs: tuple[typing.Any, ...] = tuple()
        if self.__parent_identity__ == "-":
            finalargs = args
        elif self.__parent_identity__ == "bot_class":
            finalargs = (handler.client,) + args
        elif self.__parent_identity__ == "event_handler_class":
            finalargs = (handler,) + args
        self.trigger_count += 1
        await self.callback(*finalargs)


class _InClassEventContainer:
    __identity__: str
    __internal_listeners__: dict[Event, list[EventListener]] = {}

    def __new__(cls: type[_InClassEventContainer], *args: typing.Any, **kwargs: typing.Any) -> _InClassEventContainer:
        inst = super().__new__(
            cls,
        )
        for _type in inst.__class__.__mro__:
            for item in _type.__dict__.values():
                if isinstance(item, EventListener):
                    inst.__internal_listeners__.setdefault(item.event_type, []).append(item)
                    item.__parent_identity__ = cls.__identity__
        return inst


class EventHandler(_InClassEventContainer):
    """
    Event handler to deal with incoming events from the Gateway.

    Parameters
    ----------
    client : wyvern.clients.GatewayClient
        The client binded with the event handler.

    Attributes
    ----------
    listeners: dict[str | [wyvern.events.Event], list[wyvern.events.EventListener]]
        A container for event listeners.
    """

    listeners: dict[Event, list[EventListener]] = {}
    __identity__: str = "event_handler_class"  # this is meant for callback signature verification, don't override.

    def __init__(self, client: "GatewayClient") -> None:
        self.client = client

    def setup_listeners(self) -> None:
        [[self.add_listener(listener) for listener in listeners] for listeners in self.__internal_listeners__.values()]

    def add_listener(self, event_listener: EventListener) -> EventListener:
        """Adds a listener to the handler.

        Parameters
        ----------
        event_listener: wyvern.events.EventListener
            The listener to be added.
        """
        self.listeners.setdefault(event_listener.event_type, []).append(event_listener)
        event_listener.event_handler = self
        return event_listener

    def dispatch(self, event: Event, *args: typing.Any) -> None:
        """Dispatches events from the gateway.
        This method runs all the listeners registered in the container
        for the specific event.

        Parameters
        ----------
        event: wyvern.events.Event
            Name of the event to be dispatched.
        *args: tuple[typing.Any, ...]
            Arguments to provide in callbacks.
        """
        self.client._logger.debug(f"Dispatching {event} event.")

        invokes = [lsnr(*args) for lsnr in self.listeners.get(event, []) if lsnr.max_trigger > lsnr.trigger_count]
        asyncio.gather(*invokes)


def as_listener(
    event: Event, *, max_trigger: int | float = float("inf")
) -> typing.Callable[[typing.Callable[..., typing.Awaitable[typing.Any]]], EventListener]:
    """Creates a [wyvern.events.EventListener][] object.

    Parameters
    ----------
    event: wyvern.events.Event
        The event to listen.
    max_trigger: int | float
        Maximum number of times this listener has to be triggered.

    Returns
    -------
    wyvern.events.EventListener
        A EventListener object.

    """

    def inner(callback: typing.Callable[..., typing.Awaitable[typing.Any]]) -> EventListener:
        nonlocal event, max_trigger
        return EventListener(event, callback, max_trigger, 0)

    return inner



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import datetime
import typing

import attrs

from wyvern.files import Avatar, AvatarType

from .base import DiscordObject, Snowflake

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient

__all__: tuple[str, ...] = ("User", "BotUser")


@attrs.define(kw_only=True, slots=True, repr=True)
class User(DiscordObject):
    """Represents a discord user."""

    client: "GatewayClient"

    raw: dict[str, typing.Any]
    id: Snowflake
    """ID of the user."""
    username: str
    """Username of the user."""
    discriminator: str
    """User's discriminator."""
    avatar_hash: str | None
    """Hash of the user's avatar"""
    is_bot: bool
    """True if the user is a bot."""
    is_system: bool
    """True if the user is a system user."""
    is_mfa_enabled: bool
    """True if the user has mfa enabled ( can be false even if it's enabled )"""
    banner_hash: str | None
    """User's banner hash."""
    accent_color: int | None
    """User's accent color."""
    locale: str | None
    """User's locale."""
    flags_value: int | None
    """Integer value for user flags"""
    premium_type_value: int | None
    """Premium type integer."""
    public_flags_value: int | None
    """Integer value for user's public flags"""

    def __str__(self) -> str:
        return f"{self.username}#{self.discriminator}"

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which user was created."""
        return self.get_created_at(self.id)

    @property
    def avatar(self) -> Avatar | None:
        return Avatar(client=self.client, type=AvatarType.CUSTOM, hash=self.avatar_hash) if self.avatar_hash else None

    @property
    def default_avatar(self) -> Avatar:
        return Avatar(client=self.client, type=AvatarType.DEFAULT, hash=f"embed/avatars/{int(self.discriminator)%5}")

    @property
    def display_avatar(self) -> Avatar:
        return self.avatar or self.default_avatar


@typing.final
class BotUser(User):
    """Represents the bot user on discord."""

    async def edit(self, username: str | None = None, avatar: bytes | None = None) -> "BotUser":
        """Edits the bot's user.

        Arguments
        ---------

        username : str
            The new username.
        avatar : bytes
            The new avatar bytes.

        Returns
        -------

        wyvern.models.users.BotUser
            The updated user of bot.
        """
        return await self.client.rest.edit_client_user(username, avatar)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime
import typing

import attrs

from .base import DiscordObject, Snowflake

if typing.TYPE_CHECKING:
    from .guilds import Guild

__all__: tuple[str, ...] = (
    "Role",
    "RoleTags",
)


@attrs.define(kw_only=True, slots=True, repr=True)
class RoleTags(DiscordObject):
    raw: dict[str, Role]
    bot_id: typing.Optional[Snowflake]
    integration_id: typing.Optional[Snowflake]
    premium_subscriber: typing.Optional[bool]

    def is_bot_managed(self) -> bool:
        return self.bot_id is not None

    def is_integration_managed(self) -> bool:
        return self.integration_id is not None

    def is_premium_subscriber(self) -> bool:
        return self.premium_subscriber is not None

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which role was created."""
        return super().created_at


@attrs.define(kw_only=True, slots=True, repr=True, order=False)
class Role(DiscordObject):
    raw: dict[str, typing.Any]
    id: Snowflake
    name: str
    color: int
    is_hoisted: bool
    icon: typing.Optional[str]
    unicode_emote: typing.Optional[str]
    position: int
    permissions: str
    is_managed: bool
    is_mentionable: bool
    tags: typing.Optional[RoleTags]
    guild: "Guild"

    def __lt__(self, other: typing.Any) -> bool:
        if not isinstance(other, Role) or not isinstance(self, Role):
            return NotImplemented
        if self.guild != other.guild:
            raise ValueError("Cannot compare roles from different guilds.")
        # everyone role is always at the bottom
        if self.id == self.guild.id:
            return other.id != self.guild.id
        if self.position < other.position:
            return True
        if self.position == other.position:
            return int(self.id) > int(other.id)
        return False

    def __le__(self, other: Role) -> bool:
        r = Role.__lt__(self, other)
        if r is NotImplemented:
            return NotImplemented
        return not r

    def __gt__(self, other: Role) -> bool:
        return Role.__le__(self, other)

    def __ge__(self, other: Role) -> bool:
        r = Role.__lt__(self, other)
        if r is NotImplemented:
            return NotImplemented
        return not r

    def is_bot_managed(self) -> bool:
        return self.tags.is_bot_managed() if self.tags else False

    def is_integration_managed(self) -> bool:
        return self.tags.is_integration_managed() if self.tags else False

    def is_premium_subscriber(self) -> bool:
        return self.tags.is_premium_subscriber() if self.tags else False

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which role was created."""
        return super().created_at



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

import attrs

from wyvern import files

from .base import Interaction, InteractionCommandOptionType, InteractionCommandType, InteractionType

if typing.TYPE_CHECKING:
    from wyvern import models
    from wyvern.commands.slash_commands import CommandOption
    from wyvern.permissions import Permissions

    from .localizations import Localizations


__all__: tuple[str, ...] = (
    "ApplicationCommandInteraction",
    "ApplicationCommandInteractionData",
    "InteractionOption",
    "AutocompleteInteraction",
    "ApplicationCommandInteractionResolvedData",
)


@typing.final
@attrs.define(kw_only=True, slots=True)
class InteractionOption:
    """Represents a interaction option for slash commands. (From the api.)

    !!! note
        This class is not the same as [commands.slash_commands.CommandOption][] class.
    """

    name: str
    """Name of the option."""
    type: InteractionCommandOptionType
    """Type of the option."""
    value: int | str | None
    """Value of the option."""
    options: list[InteractionOption]
    """List of options in the options, if any."""
    required: bool = True
    """True if the option is required."""
    is_focused: bool
    """If the current option focused in an autocomplete interaction."""
    autocomplete: bool = False
    """Weather to enable autocomplete for this option or not."""
    choices: list[CommandOption] = []
    """List of choices to add to the option."""
    channel_types: list[models.ChannelType] = []
    """List of channel types for channel options."""
    min_value: int | None = None
    """Minimum value for the option"""
    max_value: int | None = None
    """Maximum value for the option"""
    min_length: int | None = None
    """Minimum length acceptable for the option"""
    max_length: int | None = None
    """Maximum length acceptable for the option"""
    name_locales: Localizations | None = None
    """Name locales for the option description"""
    description_locales: Localizations | None = None
    """Name locales for the option description"""

    def __str__(self) -> str:
        return self.name


@typing.final
@attrs.define(kw_only=True, slots=True)
class ApplicationCommandInteractionResolvedData:
    """Resloved data of the app command interaction."""

    users: dict[models.base.Snowflake, models.User]
    members: dict[models.base.Snowflake, models.Member]
    messages: dict[models.base.Snowflake, models.Message]
    attachments: dict[models.base.Snowflake, files.Attachment]
    channels: dict[models.base.Snowflake, models.ChannelLike]


@typing.final
@attrs.define(kw_only=True, slots=True, repr=True)
class ApplicationCommandInteractionData:
    """Data related to an application command interaction."""

    payload: dict[str, typing.Any]
    """Raw payload recieved from discord."""
    type = InteractionType.APPLICATION_COMMAND
    """Type of the interaction."""
    command_id: models.base.Snowflake
    """ID of the command."""
    command_name: str
    """Name of the command."""
    command_type: InteractionCommandType
    """Type of the interaction command."""
    guild_id: models.base.Snowflake | None
    """ID of the guild where interaction was triggered"""
    target_id: models.base.Snowflake | None
    """ID of the application command target."""
    options: list[InteractionOption]
    """List of options used in the interaction."""
    resoloved: ApplicationCommandInteractionResolvedData | None
    """The resolved interaction data."""


class ApplicationCommandInteraction(Interaction):
    """Represents an application command interaction."""

    data: ApplicationCommandInteractionData


class AutocompleteInteraction(Interaction):
    """Represents an autocomplete interaction."""

    ...


@attrs.define(kw_only=True)
class APIApplicationCommand:
    id: models.base.Snowflake
    type: InteractionCommandType = InteractionCommandType.MESSAGE
    application_id: models.base.Snowflake
    guild_id: models.base.Snowflake | None = None
    name: str
    name_locales: Localizations | None = None
    description: str
    description_locales: Localizations | None = None
    options: InteractionOption
    default_member_permissions: Permissions | None = None
    dm_permissions: bool = False
    version: models.base.Snowflake



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

import attrs

from wyvern._internals import BitWiseFlag

if typing.TYPE_CHECKING:
    from wyvern.models.base import Snowflake

__all__: tuple[str, ...] = ("Permissions", "PermissionOverwrites")


class Permissions(BitWiseFlag):
    CREATE_INVITE = 1 << 0
    KICK_MEMBERS = 1 << 1
    BAN_MEMBERS = 1 << 2
    ADMINISTRATOR = 1 << 3
    MANAGE_CHANNELS = 1 << 4
    MANAGE_GUILD = 1 << 5
    ADD_REACTIONS = 1 << 6
    VIEW_AUDIT_LOG = 1 << 7
    PRIORITY_SPEAKER = 1 << 8
    STREAM = 1 << 9
    VIEW_CHANNEL = 1 << 10
    SEND_MESSAGES = 1 << 11
    SEND_TTS_MESSAGES = 1 << 12
    MANAGE_MESSAGES = 1 << 13
    EMBED_LINKS = 1 << 14
    ATTACH_FILES = 1 << 15
    READ_MESSAGE_HISTORY = 1 << 16
    MENTION_EVERYONE = 1 << 17
    USE_EXTERNAL_EMOJIS = 1 << 18
    VIEW_GUILD_INSIGHTS = 1 << 19
    CONNECT = 1 << 20
    SPEAK = 1 << 21
    MUTE_MEMBERS = 1 << 22
    DEAFED_MEMBERS = 1 << 23
    MOVE_MEMBERS = 1 << 25
    CHANGE_NICKNAME = 1 << 26
    MANAGE_NICKNAMES = 1 << 27
    MANAGE_ROLES = 1 << 28
    MANAGE_WEBHOOKS = 1 << 29
    MANAGE_EMOJIS_AND_STICKERS = 1 << 30
    USE_APPLICATION_COMMANDS = 1 << 31
    REQUEST_TO_SPEAK = 1 << 32
    MANAGE_EVENTS = 1 << 33
    MANAGE_THREADS = 1 << 34
    CREATE_PUBLIC_THREADS = 1 << 35
    CREATE_PRIVATE_THREADS = 1 << 36
    USE_EXTERNAL_STICKERS = 1 << 37
    SEND_MESSAGES_IN_THREADS = 1 << 38
    USE_EMBEDDED_ACTIVITIES = 1 << 39
    MODERATE_MEMBERS = 1 << 40


class OverwriteType:
    ROLE = 0
    MEMBER = 1


@attrs.define(kw_only=True, slots=True)
class PermissionOverwrites:
    id: Snowflake
    type: int
    allow: Permissions
    deny: Permissions



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import enum
import typing

__all__: tuple[str, ...] = ("Status", "Activity", "ActivityType")


@typing.final
class Status(enum.Enum):
    """Enum for User's status"""

    ONLINE = "online"
    """Online status."""
    DND = "dnd"
    """DND status."""
    IDLE = "idle"
    """IDLE status."""
    INVISIBLE = "invisible"
    """Invisible status."""
    OFFLINE = "offline"
    """Offline status."""


@typing.final
class ActivityType(enum.IntEnum):
    """Type of activity."""

    GAME = 0
    """The user is playing a game."""
    STREAMING = 1
    """Activity where user is streaming."""
    LISTENING = 2
    """Listening activity."""
    WATCHING = 3
    """Watching activity."""
    CUSTOM = 4
    """This activity is a custom status."""
    COMPETING = 5
    """Compoeting activity."""


class Activity:
    """Represents a base activity with minumum info.
    Base class for all activities and for bot's gateway payload.

    Parameters
    ----------
    name : str
        Name of the activity.
    type : wyvern.presences.ActivityType
        Type of the activity.
    url : The URL this activity points to.
    """

    name: str
    """Name of the activity."""
    type: ActivityType
    """Type of activity."""
    url: str | None
    """URL this activity points to."""

    def __init__(self, *, name: str, type: ActivityType, url: str | None = None) -> None:
        self.name = name
        self.type = type
        self.url = url

    def to_event_payload(
        self,
    ) -> dict[str, str | int | None]:
        return {"name": self.name, "type": int(self.type), "url": self.url}



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import datetime
import typing

from wyvern.files import Attachment
from wyvern.permissions import PermissionOverwrites

from . import channels
from .base import Snowflake
from .members import Member
from .messages import Message, MessageReference
from .users import BotUser, User

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient


__all__: tuple[str, ...] = ("payload_to_user", "payload_to_member", "payload_to_botuser", "payload_to_message")


def payload_to_user(client: "GatewayClient", payload: dict[str, typing.Any]) -> "User":
    return User(
        client=client,
        raw=payload,
        id=Snowflake.create(payload["id"]),
        username=payload["username"],
        discriminator=payload["discriminator"],
        avatar_hash=payload.get("avatar"),
        is_bot=payload.get("bot", False),
        is_system=payload.get("system", False),
        is_mfa_enabled=payload.get("mfa_enabled", False),
        banner_hash=payload.get("banner"),
        accent_color=payload.get("accent_color"),
        locale=payload.get("locale"),
        flags_value=payload.get("flags"),
        premium_type_value=payload.get("premium_type"),
        public_flags_value=payload.get("public_flags"),
    )


def payload_to_member(client: "GatewayClient", guild_id: Snowflake, payload: dict[str, typing.Any]) -> Member:
    user = payload_to_user(client, payload=payload["user"])
    return Member(
        client=client,
        raw=payload,
        user=user,
        id=user.id,
        username=user.username,
        discriminator=user.discriminator,
        avatar_hash=user.avatar_hash,
        is_bot=user.is_bot,
        is_system=user.is_system,
        is_mfa_enabled=user.is_mfa_enabled,
        banner_hash=user.banner_hash,
        accent_color=user.accent_color,
        locale=user.locale,
        flags_value=user.flags_value,
        premium_type_value=user.premium_type_value,
        public_flags_value=user.public_flags_value,
        guild_id=guild_id,
        nickname=payload.get("nick"),
        premium_since=datetime.datetime.fromisoformat(p) if (p := payload.get("premium_since")) else None,
        role_ids=[Snowflake.create(id) for id in payload["roles"]],
        joined_at=datetime.datetime.fromisoformat(payload["joined_at"]),
        deaf=payload.get("deaf"),
        mute=payload.get("mute"),
        communication_disabled_until=datetime.datetime.fromisoformat(cdu)
        if (cdu := payload.get("communication_disabled_until"))
        else None,
        pending=payload.get("pending", None),
        guild_avatar_hash=payload.get("avatar"),
    )


def payload_to_botuser(client: "GatewayClient", payload: dict[str, typing.Any]) -> BotUser:
    return BotUser(
        client=client,
        raw=payload,
        id=Snowflake.create(payload["id"]),
        username=payload["username"],
        discriminator=payload["discriminator"],
        avatar_hash=payload.get("avatar"),
        is_bot=payload.get("bot", False),
        is_system=payload.get("system", False),
        is_mfa_enabled=payload.get("mfa_enabled", False),
        banner_hash=payload.get("banner"),
        accent_color=payload.get("accent_color"),
        locale=payload.get("locale"),
        flags_value=payload.get("flags"),
        premium_type_value=payload.get("premium_type"),
        public_flags_value=payload.get("public_flags"),
    )


def payload_to_message(client: "GatewayClient", payload: dict[str, typing.Any]) -> Message:
    return Message(
        client=client,
        raw=payload,
        id=Snowflake.create(payload["id"]),
        tts=payload["tts"],
        pinned=payload["pinned"],
        mentions=[payload_to_user(client, item) for item in payload["mentions"]],
        mention_roles=payload["mention_roles"],
        flags_value=payload["flags"],
        embeds=[],
        edited_at=datetime.datetime.fromisoformat(t) if (t := payload.get("edited_timestamp")) else None,
        content=payload["content"],
        channel_id=payload["channel_id"],
        author=payload_to_user(client, payload["author"]),
        attachments=[Attachment._from_payload(client, Snowflake, attc) for attc in payload["attachments"]],
        message_reference=MessageReference(**data) if (data := payload.get("message_reference")) else None,
    )


def payload_to_channel(client: GatewayClient, payload: dict[str, typing.Any]) -> channels.ChannelLike:
    channel: channels.ChannelLike
    if payload["type"] == channels.ChannelType.GUILD_TEXT.value:
        channel = channels.GuildTextChannel(
            client=client,
            raw=payload,
            id=Snowflake.create(payload["id"]),
            name=payload["name"],
            guild_id=Snowflake.create(payload["guild_id"]),
            position=payload["position"],
            permissions_overwrites=[PermissionOverwrites(**perm) for perm in payload.get("permission_overwrites", [])],
            rate_limit_per_user=payload["rate_limit_per_user"],
            nsfw=payload.get("nsfw", False),
            topic=payload.get("topic", None),
            last_message_id=Snowflake.create(id) if (id := payload.get("last_message_id")) else None,
            parent_id=Snowflake.create(id) if (id := payload.get("parent_id")) else None,
            default_auto_archive_duration=payload["default_auto_archive_duration"],
        )
    elif payload["type"] == channels.ChannelType.GUILD_VOICE.value:
        channel = channels.GuildVoiceChannel(
            client=client,
            raw=payload,
            id=Snowflake.create(payload["id"]),
            name=payload["name"],
            guild_id=Snowflake.create("guild_id"),
            position=payload["position"],
            permissions_overwrites=[PermissionOverwrites(**perm) for perm in payload.get("permission_overwrites", [])],
            nsfw=payload.get("nsfw", False),
            parent_id=Snowflake.create(id) if (id := payload.get("parent_id")) else None,
            rate_limit_per_user=payload["rate_limit_per_user"],
            rtc_region=payload.get("rtc_region"),
            bitrate=payload["birtate"],
            user_limit=payload["user_limit"],
        )
    elif payload["type"] == channels.ChannelType.DM.value:
        channel = channels.DMChannel(
            client=client,
            raw=payload,
            id=Snowflake.create(payload["id"]),
            last_message_id=Snowflake.create(id) if (id := payload.get("last_message_id")) else None,
            recipients=[payload_to_user(client, user) for user in payload["recipients"]],
        )
    elif payload["type"] in (channels.ChannelType.PUBLIC_THREAD.value, channels.ChannelType.PRIVATE_THREAD.value):
        channel = channels.ThreadChannel(
            client=client,
            raw=payload,
            type=channels.ChannelType.PUBLIC_THREAD if payload["type"] == 11 else channels.ChannelType.PRIVATE_THREAD,
            id=Snowflake.create(payload["id"]),
            name=payload["name"],
            guild_id=Snowflake.create(payload["guild_id"]),
            permissions_overwrites=[PermissionOverwrites(**perm) for perm in payload.get("permission_overwrites", [])],
            parent_id=payload["parent_id"],
            owner_id=payload["owner_id"],
            last_message_id=Snowflake.create(id) if (id := payload.get("last_message_id")) else None,
            position=payload["position"],
            message_count=payload["message_count"],
            rate_limit_per_user=payload["rate_limit_per_user"],
            total_message_sent=payload["total_message_sent"],
            member_count=payload["member_count"],
        )
    elif payload["type"] == channels.ChannelType.GUILD_CATEGORY.value:
        channel = channels.GuildCategoryChannel(
            client=client,
            raw=payload,
            name=payload["name"],
            id=Snowflake.create(payload["id"]),
            guild_id=Snowflake.create(payload["id"]),
            position=payload["position"],
            permissions_overwrites=[PermissionOverwrites(**perm) for perm in payload.get("permission_overwrites", [])],
        )
    elif payload["type"] == channels.ChannelType.ANNOUNCEMENT_THREAD.value:
        channel = channels.ThreadAnnouncementChannel(
            client=client,
            raw=payload,
            id=Snowflake.create(payload["id"]),
            name=payload["name"],
            guild_id=Snowflake.create(payload["guild_id"]),
            permissions_overwrites=[PermissionOverwrites(**perm) for perm in payload.get("permission_overwrites", [])],
            parent_id=payload["parent_id"],
            owner_id=payload["owner_id"],
            last_message_id=Snowflake.create(id) if (id := payload.get("last_message_id")) else None,
            position=payload["position"],
            message_count=payload["message_count"],
            rate_limit_per_user=payload["rate_limit_per_user"],
            total_message_sent=payload["total_message_sent"],
            member_count=payload["member_count"],
        )
    elif payload["type"] == channels.ChannelType.GUILD_ANNOUNCEMENT.value:
        channel = channels.GuildAnnouncmentChannel(
            client=client,
            raw=payload,
            id=Snowflake.create(payload["id"]),
            name=payload["name"],
            guild_id=Snowflake.create(payload["guild_id"]),
            position=payload["position"],
            permissions_overwrites=[PermissionOverwrites(**perm) for perm in payload.get("permission_overwrites", [])],
            nsfw=payload.get("nsfw", False),
            topic=payload.get("topic", None),
            last_message_id=Snowflake.create(id) if (id := payload.get("last_message_id")) else None,
            parent_id=Snowflake.create(id) if (id := payload.get("parent_id")) else None,
            default_auto_archive_duration=payload["default_auto_archive_duration"],
        )
    elif payload["type"] == channels.ChannelType.GUILD_STAGE_VOICE.value:
        channel = channels.GuildStageVoiceChannel(
            client=client,
            raw=payload,
            id=Snowflake.create(payload["id"]),
            name=payload["name"],
            permissions_overwrites=[PermissionOverwrites(**perm) for perm in payload.get("permission_overwrites", [])],
            guild_id=payload["guild_id"],
            position=payload["position"],
            nsfw=payload.get("nsfw", False),
        )
    else:
        raise ValueError("Unidenfitied channel type: %s", payload["type"])
    return channel



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import datetime
import typing

import attrs

from wyvern._internals import BitWiseFlag

from .base import DiscordObject, Snowflake

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient
    from wyvern.components.container import ActionRowContainer
    from wyvern.constructors.embeds import Embed, EmbedConstructor
    from wyvern.files import Attachment

    from .users import User

__all__: tuple[str, ...] = ("MessageReference", "MessageFlags", "Message", "AllowedMentions")


@attrs.define(kw_only=True, slots=True, repr=True)
class MessageReference:
    channel_id: int | None = None
    """ID of the channel."""
    message_id: int | None = None
    """ID of the message."""
    guild_id: int | None = None
    """ID of the guild, if any."""
    fail_if_not_exists: bool = True
    """Weather to raise error if message does not exist."""

    def to_payload(self) -> dict[str, int | bool | None]:
        """Converts the object to a sendable payload.

        Returns
        -------

        dict[str, int| bool]
        """
        return {
            "channel_id": self.channel_id,
            "message_id": self.message_id,
            "guild_id": self.guild_id,
            "fail_if_not_exists": self.fail_if_not_exists,
        }


@attrs.define(kw_only=True, slots=True)
class AllowedMentions:
    """Represents an allowed mentions payload for discord."""

    roles: bool = False
    """Set to True if role mentions are allowed."""
    users: bool = False
    """Set to True if user mentions are allowed."""
    everyone: bool = False
    """Set to True if `@everyone` mentions are allowed."""
    replied_user: bool = False
    """Set to True if bot should ping message authors on replies."""

    def to_payload(self) -> dict[str, list[str] | bool]:
        payload: dict[str, list[str] | bool] = {"parse": [], "replied_user": self.replied_user}
        assert isinstance(payload["parse"], list)
        if self.roles is True:
            payload["parse"].append("roles")
        if self.users is True:
            payload["parse"].append("users")
        if self.everyone is True:
            payload["parse"].append("everyone")

        return payload


class MessageFlags(BitWiseFlag):
    CROSSPOSTED = 1 << 0
    IS_CROSSPOST = 1 << 1
    SUPPRESS_EMBEDS = 1 << 2
    SOURCE_MESSAGE_DELETED = 1 << 3
    URGENT = 1 << 4
    HAS_THREAD = 1 << 5
    EPHEMERAL = 1 << 6
    LOADING = 1 << 7
    FAILED_TO_MENTION_SOME_ROLES_IN_THREAD = 1 << 8


@attrs.define(kw_only=True, slots=True, repr=True, eq=True)
class Message(DiscordObject):
    """
    Represents a discord message.
    """

    client: "GatewayClient"

    raw: dict[str, typing.Any]
    id: Snowflake
    """ID of the message."""
    tts: bool
    """True if the message was a TTS message."""
    pinned: bool
    """Is the message is pinned."""
    mentions: list["User"]
    """List of users mentioned in the message."""
    mention_roles: list[int]
    """List of ids for roles mentioned in the message."""
    flags_value: int
    """Value of message flags"""
    embeds: list["Embed"]
    """List of embeds attached to the message."""
    edited_at: datetime.datetime | None
    """When was the message last edited."""
    content: str | None
    """The message content.
    Needs [wyvern.intents.Intents.MESSAGE_CONTENT] to be access this in a guild context."""
    channel_id: int
    """ID of the channel where this message was sent."""
    author: "User"
    """The message author."""
    message_reference: MessageReference | None
    """The message reference, if any."""
    attachments: list[Attachment]
    """File attachments in this message."""

    async def respond(
        self,
        content: str | None = None,
        *,
        embed: "EmbedConstructor" | None = None,
        embeds: typing.Sequence["EmbedConstructor"] = (),
        components: typing.Sequence["ActionRowContainer"] = (),
        allowed_mentions: AllowedMentions | None = None,
        flags: MessageFlags | None = None,
        delete_after: int | None = None,
        reply: bool = False,
    ) -> "Message":
        """
        Parameters
        ----------

        content : str | None
            The text content of the message.
        embed : wyvern.constructors.embeds.EmbedConstructor
            An single embed object to send, for multiple use `embeds` kwarg instead.
        embeds : typing.Sequence[wyvern.constructors.embeds.EmbedConstructor]
            Sequence of embeds to send.
        components : typing.Sequence[wyvern.components.container.ActionRowContainer]
            Sequence of action rows to send.
        allowed_mentions : wyvern.models.messages.AllowedMentions | None
            Allowed mentions configs.
        reply : bool
            True if you the response should be created a reply.

        Returns
        -------

        wyvern.models.messages.Message
            The message object that got created.
        """

        if all([embed, embeds]):
            raise ValueError("You cannot use both embed and embeds kwarg.")
        embeds = [embed] if embed else embeds
        return await self.client.rest.create_message(
            self.channel_id,
            content,
            embeds=embeds,
            components=components,
            allowed_mentions=allowed_mentions,
            reference=self.id if reply is True else None,
            delete_after=delete_after,
            flags=flags,
        )



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import enum
import logging
import typing

__all__: tuple[str, ...] = ("LoggingFormatter", "create_logging_setup", "ANSI", "ANSIBuilder")


class LoggingFormatter(logging.Formatter):

    COLOR_CONFIGS = {
        logging.DEBUG: "\x1b[33;49m",
        logging.INFO: "\x1b[32;49m",
        logging.WARNING: "\x1b[35;49m",
        logging.ERROR: "\x1b[31;49m",
        logging.CRITICAL: "\x1b[33;41;1m",
    }

    def format(self, record: logging.LogRecord) -> str:
        log_format = f"[%(asctime)s : {record.levelname.rjust(7)}] | %(message)s "

        formatter = logging.Formatter(
            "".join((self.COLOR_CONFIGS.get(record.levelno, "\x1b[32;49m"), log_format, "\x1b[0m"))
        )
        return formatter.format(record)


def create_logging_setup(logger: logging.Logger) -> None:
    stream = logging.StreamHandler()
    stream.setFormatter(LoggingFormatter())
    logger.addHandler(stream)
    logger.setLevel(logging.INFO)


class ANSI(enum.IntEnum):
    NORMAL_FORMAT = 0
    BOLD_FORMAT = 1
    UNDERLINE_FORMAT = 4
    GRAY_TEXT = 30
    RED_TEXT = 31
    GREEN_TEXT = 32
    YELLOW_TEXT = 33
    BLUE_TEXT = 34
    PINK_TEXT = 35
    CYAN_TEXT = 36
    WHITE_TEXT = 37
    FIREFLY_DARK_BLUE_BACKGROUND = 40
    ORANGE_BACKGROUND = 41
    MARBLE_BLUE_BACKGROUND = 42
    GREYISH_TURQUOISE_BACKGROUND = 43
    GRAY_BACKGROUND = 44
    INDIGO_BACKGROUND = 45
    LIGHT_GRAY_BACKGROUND = 46
    WHITE_BACKGROUND = 0


class ANSIBuilder:
    bucket: list[str]
    current_cursor: str

    def __enter__(self) -> ANSIBuilder:
        self.bucket = []
        self.current_cursor = ""
        return self

    def __exit__(self, *args: typing.Any) -> None:
        ...

    def set_cursor(self, *args: ANSI | int) -> ANSIBuilder:
        self.current_cursor = (
            f"\033[{';'.join(map(lambda arg: str(arg.value) if isinstance(arg, ANSI) else str(arg), args))}m"
        )
        self.bucket.append(self.current_cursor)
        return self

    def write(self, text: str, *, newline: bool = False) -> ANSIBuilder:
        self.bucket.append(text + "\n" if newline is True else "")

        return self

    def reset(self) -> ANSIBuilder:
        self.current_cursor = "\033[0m"
        self.bucket.append(self.current_cursor)
        return self

    def get_str(self) -> str:
        return "".join(self.bucket)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import enum
import typing

import attrs

from .base import Component, ComponentType

__all__: tuple[str, ...] = ("TextInputStyle", "TextInput", "Modal")


class TextInputStyle(enum.IntFlag):
    SHORT = 1
    """For single line text-inputs."""
    PARAGRAPH = 2
    """For multi line text-inputs."""


@typing.final
@attrs.define(kw_only=True, slots=True, repr=True)
class TextInput(Component):
    """Represents a modal text-input."""

    type: ComponentType = ComponentType.TEXT_INPUT
    custom_id: str
    """Custom id of the textinput."""
    label: str
    """Label of the textinput."""
    style: TextInputStyle = TextInputStyle.SHORT
    """Textinput style."""
    min_length: int | None = None
    """The maximum allowed length for textinput."""
    max_length: int | None = None
    """The minimum allowed length for textinput."""
    required: bool = True
    """Weather the field is required."""
    default_value: str | None = None
    """Default value for this textinput."""
    placeholder: str | None = None
    """The placeholder to use, if any."""

    def to_payload(self) -> dict[str, typing.Any]:
        return {
            "type": int(self.type),
            "custom_id": self.custom_id,
            "label": self.label,
            "style": int(self.style),
            "min_length": self.min_length,
            "max_length": self.max_length,
            "required": self.required,
            "value": self.default_value,
            "placeholder": self.placeholder,
        }


@attrs.define(kw_only=True, slots=True)
class Modal:
    """Represents a discord modal form."""

    title: str
    """Title of the modal."""
    custom_id: str
    """Custom ID of the modal"""
    text_inputs: list[TextInput] = []
    """A list of [TextInputs] to use in the modal."""

    def add_item(self, text_input: TextInput) -> None:
        """Method to add a textinput to the Modal."""
        self.text_inputs.append(text_input)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import builtins
import importlib
import inspect
import logging
import typing

# isort: off
from wyvern import (
    commands,
    events,
    exceptions,
    gateway,
    interactions,
    models,
    presences,
    rest,
    state_handlers,
    utils,
    ux,
)

# isort: on
from wyvern import intents as _intents
from wyvern import plugins as _plugins

if typing.TYPE_CHECKING:
    import aiohttp

    from wyvern._types import AppCommandCallbackT


_LOGGER = logging.getLogger(__name__)

ux.create_logging_setup(_LOGGER)

__all__: tuple[str, ...] = ("GatewayClient", "CommandsClient")


class GatewayClient(events._InClassEventContainer):
    """The main bot class which acts as an interface between the Discord API and your bot.

    Parameters
    ----------
        token : str
        The bot token to use.
    intents : typing.SupportsInt | wyvern.intents.Intents
        The intents to use while logging in to the gateway.
    allowed_mentions : wyvern.models.messages.AllowedMentions
        The default mentions to allow in a message bot is sending.
    event_handler type[EventHandler]
        A EventHandler subclass ( not instance ), if any.
    rest_client : RESTClient | None
        A custom RESTClient subclass to use, if any.
    api_version : int
        Discord API version to use.
    client_session : aiohttp.ClientSession | None
        ClientSession subclass to use, if any.
    """

    __identity__: str = "bot_class"
    hooks: dict[str, utils.Hook] = {}
    """Mapping of loaded hook names to their objects."""
    plugins: dict[str, _plugins.Plugin] = {}
    """Mapping of loaded plugin names to their objects."""

    def __init__(
        self,
        token: str,
        *,
        intents: typing.SupportsInt | _intents.Intents = _intents.Intents.UNPRIVILEGED,
        event_handler: type[events.EventHandler] = events.EventHandler,
        allowed_mentions: models.AllowedMentions = models.messages.AllowedMentions(),
        rest_client: rest.RESTClient | None = None,
        api_version: int = 10,
        client_session: aiohttp.ClientSession | None = None,
    ) -> None:
        self._client_id: int = 0
        self.event_handler = event_handler(self)
        self.rest = rest_client or rest.RESTClient(
            client=self, token=token, api_version=api_version, client_session=client_session
        )
        self.intents = intents if isinstance(intents, _intents.Intents) else _intents.Intents(int(intents))
        self.gateway = gateway.Gateway(self)
        self.allowed_mentions = allowed_mentions
        self._users = state_handlers.UsersState(self)
        self._members = state_handlers.MembersState(self)
        self._logger = _LOGGER

    @property
    def users(self) -> state_handlers.UsersState:
        """The state handler for users stored in the bot's cache,
        Can also be used to perform fetch operations and parsing users from string.

        Returns
        -------
                wyvern.state_handlers.UsersState
            The handler.
        """
        return self._users

    @property
    def members(self) -> state_handlers.MembersState:
        """The member state for members in bot's cache. Similar to `users` cache.

        Returns
        -------
        wyvern.state_handlers.MembersState
                The member handler.
        """
        return self._members

    @property
    def latency(self) -> float:
        """The heartbeat latency of the gateway connection.

                Returns
                -------
        float
            The latency.
        """
        return self.gateway.latency

    async def wait_for(
        self, event: events.Event, *, timeout: int = 180, check: typing.Callable[..., bool] = lambda *_: True
    ) -> typing.Any:
        ...

    def with_listener(
        self, event: events.Event, *, max_trigger: int | float = float("inf")
    ) -> typing.Callable[[typing.Callable[..., typing.Awaitable[typing.Any]]], events.EventListener]:
        """Creates and adds a new listenet to the client's event handler.

        Parameters
        ----------
            event: str | wyvern.events.Event
            The event to listen.
        max_trigger: int | float
            Maximum number of times this listener has to be triggered.

        Returns
        -------
                wyvern.events.EventListener
            A EventListener object.

        ??? example

            ```py
            import wyvern

            client = wyvern.GatewayClient("TOKEN")


            @client.with_listener(wyvern.Event.MESSAGE_CREATE)
            async def message_create(message: wyvern.Message) -> None:
               if message.content == ".ping":
                   await message.respond("pong")


            client.run()
            ```
        """

        def inner(callback: typing.Callable[..., typing.Awaitable[typing.Any]]) -> events.EventListener:
            lsnr = events.as_listener(event, max_trigger=max_trigger)(callback)
            self.event_handler.add_listener(lsnr)
            return lsnr

        return inner

    async def start(
        self, *, activity: presences.Activity | None = None, status: presences.Status | None = None
    ) -> None:
        """Connects the bot with gateway and starts listening to events.

        Parameters
        ----------
            activity : wyvern.presences.Activity | None
            The activity bot boots up with.
        status : wyvern.presences.Status | None
            The status bot boots up with.
        """
        self.event_handler.setup_listeners()
        self.event_handler.dispatch(events.Event.STARTING, self)
        self.gateway._start_activity = activity
        self.gateway._start_status = status
        await self.gateway._get_socket_ready()
        _LOGGER.info("Logging in with bot token.")
        try:
            res = await self.rest.fetch_client_user()
            _LOGGER.info("Logged in to the gateway with bot token.")
            _LOGGER.info("(Session info) User ID: %s, Username: %s", res.id, res.username)
            self._client_id = res.id
            self.event_handler.dispatch(events.Event.STARTED, self)
            await self.gateway.listen_gateway()
        except exceptions.Unauthorized as e:
            await self.rest._session.close()
            e.message += ", Improper token was passed."
            raise e

    def run(self, *, activity: presences.Activity | None = None, status: presences.Status | None = None) -> None:
        """A non-async method which call [wyvern.clients.GatewayClient.start][].

        Parameters
        ----------
                activity : wyvern.presences.Activity | None
            The activity bot boots up with.
        status : wyvern.presences.Status | None
            The status bot boots up with.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.start(activity=activity, status=status))

    def load_hooks(self, path: str) -> dict[str, utils.Hook]:
        """Loads hooks from the provided path

        ??? note
            The path should be a python import path.

        Parameters
        ----------
        path : str
            Path to load hooks from.

        Returns
        -------
        dict[str, wyvern.utils.Hook]
            The hooks that were loaded.
        """
        module = importlib.import_module(path)
        hooks = {hook.name: hook for hook in module.__dict__.values() if isinstance(hook, utils.Hook)}
        if any(overrided := [hook for hook in hooks.keys() if hook in self.hooks.keys()]):
            self._logger.warning("Overring loaded hooks: %s", ", ".join(overrided))
        self.hooks.update(hooks)
        return hooks

    def get_hooks(self, *, awaitable: bool = True, non_awaitable: bool = True) -> list[utils.Hook]:
        """Get's loaded hooks.

        Parameters
        ----------
        awaitable : bool
            If awaitable hooks should be added to return value.
        non_awaitable : bool
            If non-awaitable hooks should be added to return value.

        Returns
        -------
        list[wyvern.utils.Hook]
            List of the hooks.
        """
        hooks: list[utils.Hook] = []
        if awaitable is True:
            hooks.extend([hook for hook in self.hooks.values() if inspect.isawaitable(hook.callback)])
        if non_awaitable is True:
            hooks.extend([hook for hook in self.hooks.values() if not inspect.isawaitable(hook.callback)])
        return hooks

    def add_plugin(self, plugin: _plugins.Plugin) -> None:
        """Adds a plugin to the bot.

        Parameters
        ----------
        plugin : wyvern.plugins.Plugin
            The plugin to be added.
        """
        if plugin.name in self.plugins.keys():
            raise exceptions.PluginException("Plugin %s is already loaded." % (plugin.name))
        self.plugins[plugin.name] = plugin
        plugin.setup_plugin(self)


class CommandsClient(GatewayClient, commands.CommandHandler):
    """Implementation of the [wyvern.GatewayClient][] class with a command handler."""

    slash_commands: dict[str, "commands.slash_commands.SlashCommand"] = {}
    """List of slash commands attached to the client in code."""
    slash_groups: dict[str, "commands.slash_commands.SlashGroup"] = {}
    """List of slash command groups attached to the client in code."""

    @events.as_listener(events.Event.INTERACTION_CREATE)
    async def _handle_inters(self, inter: interactions.Interaction) -> None:
        if isinstance(inter, interactions.ApplicationCommandInteraction):
            await self.process_application_commands(inter)

    def add_slash_command(self, command: commands.slash_commands.SlashCommand) -> None:
        if (name := command.name) in self.slash_commands.keys():
            raise exceptions.CommandAlreadyExists(f"A slash command named {name} already exists.")
        _LOGGER.debug(f"Adding slash command {name} to bot.")
        self.slash_commands[name] = command

    def add_slash_group(self, group: commands.slash_commands.SlashGroup) -> None:
        if (name := group.name) in self.slash_groups.keys():
            raise exceptions.CommandAlreadyExists(f"A slash command named {name} already exists.")
        _LOGGER.debug(f"Adding slash group {name} to bot.")
        self.slash_groups[name] = group

    @typing.overload
    def add_item(self, listener_or_command: events.EventListener) -> events.EventListener:
        ...

    @typing.overload
    def add_item(
        self, listener_or_command: commands.slash_commands.SlashCommand
    ) -> commands.slash_commands.SlashCommand:
        ...

    def add_item(
        self, listener_or_command: events.EventListener | commands.slash_commands.SlashCommand | typing.Any
    ) -> events.EventListener | commands.slash_commands.SlashCommand:
        def inner() -> None:
            nonlocal listener_or_command
            if isinstance(listener_or_command, events.EventListener):
                self.event_handler.add_listener(listener_or_command)
            elif isinstance(listener_or_command, commands.slash_commands.SlashCommand):
                self.add_slash_command(listener_or_command)

        inner()
        return listener_or_command

    def set_prefix(self, prefix_or_function: str | typing.Sequence[str] | builtins.function) -> "CommandsClient":
        """
        Set a prefix to parse message commands.
        Allowed prefix data
        -------------------
        * strings : [str][]
        * iterables : [list][] / [set][] / [tuple][]
        * callables : [builtins.function][] or a [asyncio.coroutine][].
        """
        if isinstance(prefix_or_function, str):
            self.prefix_type = str
        elif (
            isinstance(prefix_or_function, list)
            or isinstance(prefix_or_function, tuple)
            or isinstance(prefix_or_function, set)
        ):
            self.prefix_type = type(prefix_or_function)
        elif isinstance(prefix_or_function, builtins.function):
            self.prefix_type = builtins.function
        return self

    def with_slash_command(
        self,
        *,
        name: str,
        description: str,
    ) -> typing.Callable[[AppCommandCallbackT], commands.slash_commands.SlashCommand]:
        """Creates a slash command.

        Parameters
        ----------
        name : str
            Name of the command.
        description : str
            Description of the command.

        Returns
        -------
        typing.Callable[..., commands.slash_commands.SlashCommand]
            A [wyvern.commands.slash_commands.SlashCommand][] when called.
        """

        def inner(callback: AppCommandCallbackT) -> commands.slash_commands.SlashCommand:
            cmd = commands.as_slash_command(name=name, description=description)(callback)

            self.add_slash_command(cmd._set_client(self))
            return cmd

        return inner



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime
import typing

import attrs

from .base import DiscordObject, Snowflake

if typing.TYPE_CHECKING:
    from .users import User


__all__: tuple[str, ...] = ("Sticker", "StickerItem", "StickerPack")


@attrs.define(kw_only=True, slots=True, repr=True)
class Sticker(DiscordObject):
    raw: dict[str, typing.Any]
    id: Snowflake
    pack_id: Snowflake
    name: str
    description: typing.Optional[str]
    tags: typing.Optional[str]
    asset: str
    sticker_type: int
    format_type: int
    available: bool
    guild_id: Snowflake
    user: User
    sort_value: int

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which sticker was created."""
        return super().created_at


@attrs.define(kw_only=True, slots=True, repr=True)
class StickerItem(DiscordObject):
    raw: dict[str, typing.Any]
    id: Snowflake
    name: str
    format_type: int

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which sticker was created."""
        return super().created_at


@attrs.define(kw_only=True, slots=True, repr=True)
class StickerPack(DiscordObject):
    raw: dict[str, typing.Any]
    id: Snowflake
    stickers: list[StickerItem]
    name: str
    sku_id: Snowflake
    cover_sticker_id: Snowflake
    description: str
    banner_asset_id: Snowflake

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which sticker was created."""
        return super().created_at



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wyvern',
 'wyvern.commands',
 'wyvern.components',
 'wyvern.constructors',
 'wyvern.extensions',
 'wyvern.gateway',
 'wyvern.interactions',
 'wyvern.models',
 'wyvern.rest',
 'wyvern.state_handlers']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0', 'attrs>=22.1.0,<23.0.0', 'colorama>=0.4.6,<0.5.0']

setup_kwargs = {
    'name': 'wyvern',
    'version': '0.1.1',
    'description': 'A flexible and easy to use Discord API wrapper for python 🚀.',
    'long_description': '# wyvern\n\n<p align="center">\n<img src="https://raw.githubusercontent.com/sarthhh/wyvern/master/docs/assets/wyvern.png" height=150 width=150><br><br>\n<img src="https://img.shields.io/github/license/sarthhh/wyvern?style=flat-square">\n<img src="https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square">\n<img src="https://img.shields.io/badge/%20type_checker-pyright-%231674b1?style=flat-square">\n<img src="https://img.shields.io/github/stars/sarthhh/wyvern?style=flat-square">\n<img src="https://img.shields.io/github/last-commit/sarthhh/wyvern?style=flat-square">\n<img src="https://img.shields.io/pypi/pyversions/wyvern?style=flat-square">\n<img src="https://img.shields.io/pypi/v/wyvern?style=flat-square">\n<br><br>\nA [WIP] flexible and easy to use Discord API wrapper for python 🚀.\n</p>\n\n> Warning: This library is very unstable and things might not work as expected. Feel free to create an issue.\n\n## Important Links\n\nSupport server: https://discord.gg/FyEE54u9GF\n\nDocumentation: https://sarthhh.github.io/wyvern/\n\nPYPI: https://pypi.org/project/wyvern\n\n## Installation\n```sh\n$python -m pip install git+https://github.com/sarthhh/wyvern\n```\n\n## Example Code:\n\n* CommandsClient with commands support.\n```py\nimport wyvern\n\n# creating a CommandsClient object to interaction with commands.\nclient = wyvern.CommandsClient("TOKEN")\n\n# creating a slash command using with_slash_command decorator.\n@client.with_slash_command(name="hello", description="says a hello")\nasync def hello(interaction: wyvern.ApplicationCommandInteraction) -> None:\n    # creating a response to the interaction.\n    await interaction.create_message_response("hi!")\n\n\n# running the bot.\nclient.run()\n\n```\n* Basic GatewayClient with listener. \n```py\nimport wyvern\n\n# creating a GatewayClient instance and storing it into the client variable.\n# this acts as the interface between your bot and the code.\n\nclient = wyvern.GatewayClient("TOKEN", intents=wyvern.Intents.UNPRIVILEGED | wyvern.Intents.MESSAGE_CONTENT)\n\n# creating an EventListener object and adding it to the client\'s event handler using the\n# @client.with_listener decorator. You can set the maximum amount of time this listener will get triggered using\n# the `max_trigger kwarg in the listener decorator.`\n\n\n@client.as_listener(wyvern.Event.MESSAGE_CREATE)\nasync def message_create(message: wyvern.Message) -> None:\n    """This coroutine is triggerd whenever the MESSAGE_CREATE event gets dispatched."""\n    if message.content and message.content.lower() == "!ping":\n        await message.respond("pong!")\n\n\n# runs the bot.\n\nclient.run()\n```',
    'author': 'sarthhh',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import abc
import enum
import typing

__all__: tuple[str, ...] = ("ButtonStyle", "ComponentType")


class ButtonStyle(enum.IntEnum):
    """Enums for Button style."""

    PRIMARY = 1
    """A primary blurple discord button."""
    SECONDARY = 2
    """A secondary gray discord button."""
    SUCCESS = 3
    """Green discord button."""
    DANGER = 4
    """Red discord button."""
    LINK = 5
    """Button pointing to an URL"""
    BLURPLE = PRIMARY
    """Alias for PRIMARY"""
    GRAY = SECONDARY
    """Alias for SECONDARY"""
    GREY = SECONDARY
    """Alias for SECONDARY"""
    GREEN = SUCCESS
    """Alias for SUCCESS"""
    RED = DANGER
    """Alias for DANGER"""
    URL = LINK
    """Alias for LINK"""


class ComponentType(enum.IntEnum):
    ACTION_ROW = 1
    BUTTON = 2
    STRING_SELECT = 3
    TEXT_INPUT = 4
    USER_SELECT = 5
    ROLE_SELECT = 6
    MENTIONABLE_SELECT = 7
    CHANNEL_SELECT = 8


class Component(abc.ABC):
    """Represents a discord component.
    Is the base class for other components."""

    def to_payload(self) -> dict[str, typing.Any]:
        return {}



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime
import typing

import attrs

from .base import DiscordObject, Snowflake

if typing.TYPE_CHECKING:
    from .users import User

__all__: tuple[str, ...] = (
    "VoiceState",
    "VoiceRegion",
)


@attrs.define(kw_only=True, slots=True, repr=True)
class VoiceState(DiscordObject):
    raw: dict[str, typing.Any]
    guild_id: Snowflake
    channel_id: Snowflake
    user_id: Snowflake
    member: typing.Optional[User]
    session_id: str
    deaf: bool
    mute: bool
    self_deaf: bool
    self_mute: bool
    self_stream: bool
    self_video: bool
    suppress: bool
    request_to_speak_timestamp: typing.Optional[datetime.datetime]

    @property
    def created_at(self) -> datetime.datetime:
        """Datetime at which voice state was created."""
        return super().created_at


@attrs.define(kw_only=True, slots=True, repr=True)
class VoiceRegion:
    raw: dict[str, typing.Any]
    name: str
    id: str
    vip: bool
    optimal: bool
    deprecated: bool
    custom: bool



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import typing

import attrs

from wyvern.components import ActionRowContainer, Button, ButtonStyle

if typing.TYPE_CHECKING:

    from wyvern._types import ButtonCallbackT
    from wyvern.interactions import ComponentInteraction
    from wyvern.models.emojis import CustomEmoji
    from wyvern.models.messages import Message


@attrs.define(kw_only=True)
class CallableButton(Button):
    callback: ButtonCallbackT

    def __call__(self, button: CallableButton, inter: ComponentInteraction) -> typing.Awaitable[typing.Any]:
        return self.callback(button, inter)


@attrs.define(kw_only=True)
class MessageComponents:
    timeout: int = 180
    containers: list[ActionRowContainer] = attrs.field(default=[ActionRowContainer()] * 5, init=False)
    message: Message | None = attrs.field(default=None, init=False)

    async def _timeout(self) -> None:
        if self.timeout is not None:
            await asyncio.sleep(self.timeout)
            await self.on_timeout()

    async def on_timeout(self) -> None:
        ...

    def start(self, message: Message) -> None:
        self.message = message
        asyncio.create_task(self._timeout())

    def with_button(
        self,
        *,
        style: ButtonStyle = ButtonStyle.SECONDARY,
        label: str | None = None,
        emoji: str | CustomEmoji | None = None,
        custom_id: str | None = None,
        disabled: bool = False,
        url: str | None = None,
        row: int = 0,
    ) -> typing.Callable[[ButtonCallbackT], CallableButton]:
        """Adds a button to the component handler.

        Parameters
        ----------

        style : wyvern.components.base.ButtonStyle
            The style of button.
        label : str | None
            Button's label.
        emoji : str | wyvern.models.emojis.CustomEmoji
            A unicode emoji or a custom one.
        custom_id : str
            A custom id set to the button.
        disabled : bool
            Set this to [True][] to disable clicks on button.
        url : str
            The URL this button points to.

        Returns
        -------

        wyvern.components.buttons.Button
            The button that was created.

        """

        def decorator(callback: ButtonCallbackT) -> CallableButton:
            _button = CallableButton(
                callback=callback,
                style=style,
                label=label,
                emoji=emoji,
                custom_id=custom_id,
                disabled=disabled,
                url=url,
            )
            self.containers[row].items.append(_button)
            return _button

        return decorator

    def build(self) -> list[ActionRowContainer]:
        return self.containers



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import datetime
import json
import sys
import time
import typing

import aiohttp

from wyvern.events import Event
from wyvern.interactions import _converters as inter_convertors
from wyvern.models import _converters
from wyvern.presences import Activity, Status

from .enums import WSEventEnums
from .keep_alive import KeepAlive

if typing.TYPE_CHECKING:
    from wyvern.clients import GatewayClient


__all__: tuple[str, ...] = ("Gateway",)


class Gateway:
    __slots__: typing.Tuple[str, ...] = (
        "_client",
        "_keep_alive",
        "_latency",
        "_heartbeat_interval",
        "_socket",
        "_start_activity",
        "_start_status",
        "is_connected",
    )

    def __init__(self, client: "GatewayClient") -> None:
        self._start_activity: "Activity" | None = None
        self._start_status: "Status" | None = None
        self._client = client
        self._keep_alive = KeepAlive()
        self._latency: float = 0
        self._heartbeat_interval: float = 0
        self._socket: "aiohttp.ClientWebSocketResponse"
        self.is_connected: bool = False

    @property
    def socket(self) -> "aiohttp.ClientWebSocketResponse":
        return self._socket

    @property
    def keep_alive(self) -> KeepAlive:
        return self._keep_alive

    @property
    def latency(self) -> float:
        return self._latency

    async def update_presence(self, status: Status | None = None, activity: Activity | None = None) -> None:
        payload: dict[str, typing.Any] = {
            "op": WSEventEnums.PRESECNE_UPDATE.value,
            "d": {
                "since": int(datetime.datetime.now().timestamp()),
                "afk": False,
                "status": status.value if status else "online",
                "activities": [activity.to_event_payload()] if activity else [],
            },
        }

        await self.socket.send_json(payload)

    async def update_voice_state(
        self, guild_id: int, channel_id: int, /, *, self_mute: bool = False, self_deaf: bool = False
    ) -> None:
        payload: dict[str, typing.Any] = {
            "op": WSEventEnums.VOICE_STATE_UPDATE.value,
            "d": {"guild_id": guild_id, "channel_id": channel_id, "self_mute": self_mute, "self_deaf": self_deaf},
        }
        await self.socket.send_json(payload)

    @property
    def identify_payload(self) -> dict[str, typing.Any]:
        return {
            "op": 2,
            "d": {
                "token": self._client.rest._token,
                "intents": self._client.intents.value,
                "properties": {
                    "os": sys.platform,
                    "browser": "wyvern",
                    "device": "wyvern",
                },
                "presence": {
                    "since": int(datetime.datetime.now().timestamp()),
                    "activities": [ac.to_event_payload() if isinstance(ac := self._start_activity, Activity) else {}],
                    "status": self._start_status.value if self._start_status else None,
                    "afk": False,
                },
            },
        }

    async def listen_gateway(self) -> None:
        self._client._logger.debug("Starting listening to gateway.")
        async for message in self.socket:  # type: ignore
            if self.is_connected is False:
                self.is_connected = True
                self._client.event_handler.dispatch(Event("GATEWAY_CONNECTED"), self._client)

            if message.type == aiohttp.WSMsgType.TEXT:  # type: ignore
                await self._parse_payload_response(json.loads(message.data))  # type: ignore
            elif message.type == aiohttp.WSMsgType.ERROR:  # type: ignore
                "error from the gateway!"

    async def _get_socket_ready(self) -> None:
        self._socket = await self._client.rest._create_websocket()

    async def _hello_res(self, d: dict[str, typing.Any]) -> None:
        self._client._logger.debug("Sending identify payload.")
        await self.socket.send_json(self.identify_payload)
        self._heartbeat_interval = d["heartbeat_interval"] / 1000
        loop = asyncio.get_event_loop()
        loop.create_task(self.keep_alive.start(self))

    async def _dispatch_events(self, payload: dict[str, typing.Any]) -> None:
        try:
            event = Event(t := payload["t"])
        except ValueError:
            return
        if event is Event.MESSAGE_CREATE:
            self._client.event_handler.dispatch(event, _converters.payload_to_message(self._client, payload["d"]))
        elif event is event.INTERACTION_CREATE:
            self._client.event_handler.dispatch(
                event, inter_convertors.payload_to_interaction(self._client, payload["d"])
            )
        elif t == "GUILD_CREATE":
            self._client.event_handler.dispatch(event, payload)
            self._populate_guild_cache(payload["d"])

    async def _parse_payload_response(self, payload: dict[str, typing.Any]) -> None:
        op, s, d = payload["op"], payload["s"], payload["d"]
        if op == WSEventEnums.HEARTBEAT_ACK:
            self._latency = time.perf_counter() - self.keep_alive.last_heartbeat
            return

        if op == WSEventEnums.HELLO:
            await self._hello_res(d)

        elif op == WSEventEnums.DISPATCH:
            self.keep_alive.sequence = s
            await self._dispatch_events(payload)

    def _populate_guild_cache(self, payload: dict[str, typing.Any]) -> None:
        [
            self._client.members.add_member(_converters.payload_to_member(self._client, payload["id"], member))
            for member in payload["members"]
        ]
        self._client.members.update_user_state()



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import datetime
import typing

import attrs

from wyvern import utils

__all__: tuple[str, ...] = ("EmbedAuthor", "EmbedFooter", "EmbedField", "Embed", "EmbedConstructor")


@typing.final
@attrs.define(slots=True)
class EmbedAuthor:
    """Represents an embed author."""

    name: str
    """Name of the author."""
    url: str | None = None
    """URL that the author points to."""
    icon_url: str | None = None
    """URL of the author icon"""
    proxy_icon_url: str | None = None
    """Proxy URL of icon."""

    def to_payload(self) -> dict[str, str | None]:
        """Converts the author to its payload."""
        return {"name": self.name, "url": self.url, "icon_url": self.icon_url}


@typing.final
@attrs.define(slots=True)
class EmbedFooter:
    """Represents an embed footer"""

    text: str
    """The text in the footer."""
    icon_url: str | None = None
    """URL of the footer icon."""
    proxy_icon_url: str | None = None
    """Proxy URL of the icon"""

    def to_payload(self) -> dict[str, str | None]:
        """Converts the footer to its payload."""
        return {"text": self.text, "icon_url": self.icon_url}


@typing.final
@attrs.define(slots=True)
class EmbedField:
    """Represnts an embed field."""

    name: str
    """Name of the field."""
    value: str
    """Value of the field."""
    inline: bool = True
    """Inline value for the file, defaults to True."""

    def to_payload(self) -> dict[str, str | bool]:
        return {"name": self.name, "value": self.value, "inline": self.inline}


class EmbedConstructor:
    """
    Creates an sendable discord embed.

    Parameters
    ----------

    title : str
        Name of the embed.
    description : str
        Description of the embed
    color : int
        Color of the embed.
    url : str
        URL which the title points to.
    timestamp : datetime.datetime
        Timestamp to put in the embed.

    """

    __slots__: tuple[str, ...] = ("_payload",)

    def __init__(
        self,
        *,
        title: str | None = None,
        description: str | None = None,
        color: int = 0,
        colour: int = 0,
        url: str | None = None,
        timestamp: datetime.datetime | None = None,
    ) -> None:

        self._payload: dict[str, typing.Any] = {
            "type": "rich",
            "fields": [],
            "title": title,
            "description": description,
            "color": color or colour,
            "url": url,
            "timestamp": int(timestamp.timestamp()) if isinstance(timestamp, datetime.datetime) else None,
        }

    def add_field(self, *, name: str, value: str, inline: bool = True) -> "EmbedConstructor":
        """Adds a field to the embed.

        Parameters
        ----------

        name: str
            Name of the field.
        value: str
            Value of the filed
        inline: bool
            Weather the field is inline or not, defaults to [True][]

        Returns
        -------

        wyvern.constructors.embeds.EmbedConstructor
            The constructor.
        """
        field = EmbedField(name, value, inline)
        self._payload["fields"].append(field.to_payload())
        return self

    def set_author(self, *, name: str, url: str, icon_url: str | None = None) -> "EmbedConstructor":
        """
        Set's the author for the embed.

        Parameters
        ----------

        name: str
            Name of the author field.
        url: str
            The URL author points to.
        icon_url: str
            URL for image to embed in the author.

        Returns
        -------

        wyvern.constructors.embeds.EmbedConstructor
            The constructor.
        """
        author = EmbedAuthor(name, url, icon_url)
        self._payload["author"] = author.to_payload()
        return self

    def set_footer(self, *, text: str, icon_url: str | None = None) -> "EmbedConstructor":
        """
        Set's the footer for the embed.

        Parameters
        ----------

        text: str
            The text to appear in the footer
        icon_url: str
            URL for image to embed in the footer.

        Returns
        -------

        wyvern.constructors.embeds.EmbedConstructor
            The constructor.
        """
        footer = EmbedFooter(text, icon_url)
        self._payload["footer"] = footer.to_payload()
        return self

    def edit_init(
        self,
        *,
        title: str | None | utils.Empty = utils.EMPTY,
        description: str | None | utils.Empty = utils.EMPTY,
        color: int | utils.Empty = utils.EMPTY,
        colour: int | utils.Empty = utils.EMPTY,
        url: str | None | utils.Empty = utils.EMPTY,
        timestamp: datetime.datetime | None | utils.Empty = utils.EMPTY,
    ) -> "EmbedConstructor":
        if title != utils.EMPTY:
            self._payload["title"] = title
        if description != utils.EMPTY:
            self._payload["description"] = description
        if color != utils.EMPTY:
            self._payload["color"] = color
        if colour != utils.EMPTY:
            self._payload["color"] = colour
        if url != utils.EMPTY:
            self._payload["url"] = url
        if timestamp != utils.EMPTY:
            self._payload["timestamp"] = (
                int(timestamp.timestamp()) if isinstance(timestamp, datetime.datetime) else None
            )
        return self

    def build(self) -> "Embed":
        return Embed(
            payload=self._payload,
            title=self._payload.get("title"),
            description=self._payload.get("description"),
            url=self._payload.get("url"),
            color=self._payload.get("color", 0),
            fields=[EmbedField(**kargs.to_payload()) for kargs in self._payload.get("fields", [])],
            author=self._payload.get("author"),
            footer=self._payload.get("footer"),
        )


@typing.final
@attrs.define(slots=True, kw_only=True, eq=True, repr=True)
class Embed:
    """
    Read-only Embed class returned while parsing message objects.
    To create a sendable embed, use [wyvern.constructors.embeds.EmbedConstructor][] instead.
    """

    payload: dict[str, typing.Any]
    """The raw payload."""
    title: str | None = None
    """Title of the embed."""
    description: str | None = None
    """Description fo the embed"""
    url: str | None = None
    """URL of the embed"""
    color: int = 0
    """Color of the embed."""
    timestamp: datetime.datetime | None = None
    """Timestamp of the embed."""
    colour = color
    """Alias for color."""
    fields: list[EmbedField] = []
    """List of fields related to this embed."""
    author: EmbedAuthor | None = None
    """Author of the embed."""
    footer: EmbedFooter | None = None
    """Footer of the embed."""

    def to_constructor(self) -> EmbedConstructor:
        embed = EmbedConstructor()
        embed._payload = self.payload
        return embed

    @classmethod
    def from_payload(cls, payload: dict[str, typing.Any]) -> "Embed":
        const = EmbedConstructor()
        const._payload = payload
        return const.build()



# MIT License

# Copyright (c) 2022 Sarthak

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import attrs

LOCALE_TUPLES: list[tuple[str, str, str]] = [
    ("da", "Danish", "Dansk"),
    ("de", "German", "Deutsch"),
    ("en-GB", "English UK", "English UK"),
    ("en-US", "English US", "English US"),
    ("es-ES", "Spanish", "Español"),
    ("fr", "French", "Français"),
    ("hr", "Croatian", "Hrvatski"),
    ("it", "Italian", "Italiano"),
    ("lt", "Lithuanian", "Lietuviškai"),
    ("hu", "Hungarian", "Magyar"),
    ("nl", "Dutch", "Nederlands"),
    ("no", "Norwegian", "Norsk"),
    ("pl", "Polish", "Polski"),
    ("pt-BR", "Portuguese, Brazilian", "Português do Brasil"),
    ("ro", "Romanian, Romania", "Română"),
    ("fi", "Finnish", "Suomi"),
    ("sv-SE", "Swedish", "Svenska"),
    ("vi", "Vietnamese", "Tiếng Việt"),
    ("tr", "Turkish", "Türkçe"),
    ("cs", "Czech", "Čeština"),
    ("el", "Greek", "Ελληνικά"),
    ("bg", "Bulgarian", "български"),
    ("ru", "Russian", "Pусский"),
    ("uk", "Ukrainian", "Українська"),
    ("hi", "Hindi", "हिन्दी"),
    ("th", "Thai", "ไทย"),
    ("zh-CN", "Chinese China", "中文"),
    ("ja", "Japanese", "日本語"),
    ("zh-TW", "Chinese Taiwan", "繁體中文"),
    ("ko", "Korean", "한국어"),
]


@attrs.define
class LocaleInfo:
    symbol: str
    name: str
    native_name: str


class Localizations:
    VALID_LOCALES: dict[str, LocaleInfo] = {locale[0]: LocaleInfo(*locale) for locale in LOCALE_TUPLES}

    final_data: dict[str, str] = {}

    @classmethod
    def from_dict(cls, data: dict[str, str], /) -> "Localizations":
        (inst := cls()).final_data = data
        return inst

    def add_language(self, lang: str, value: str) -> "Localizations":
        self.final_data[lang] = value
        return self

    @classmethod
    def _check_is_valid(cls, local: str) -> str:
        assert local in cls.VALID_LOCALES.keys()
        return local
